#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <qapplication.h>
#include <qpainter.h>
#include <qtimer.h>

#include "CamWidget.h"

#include "videodev.h"
#include "pwcx.h"


CCamWidget::CCamWidget(QWidget *parent, const char *name)
          : QWidget(parent, name)
{
   m_ViewSize.x = 320;
   m_ViewSize.y = 240;
   m_Frames = 10;
   m_RealSize = m_ViewSize;
   m_Offset.x = 0;
   m_Offset.y = 0;
   m_DeviceName = "/dev/video";
   m_FD = -1;
   m_UseRaw = false;
   m_Bayer = false;
   m_pInputBuffer = 0;
   m_InputLen = 0;
   m_pPWCXBuffer = malloc(60000); // large enoug for all cams
   m_pImageBuffer = 0;
   m_ImageLen = 0;
   m_Type = 0;
}

CCamWidget::~CCamWidget()
{
   free(m_pPWCXBuffer);
   m_pPWCXBuffer = 0;
   switch(m_Type)
   {
     case 645:
     case 646:
       pwcx_exit_decompress_Nala();
       break;

     case 675:
     case 680:
     case 690:
       pwcx_exit_decompress_Timon();
       break;

     case 720:
     case 730:
     case 740:
     case 750:
       pwcx_exit_decompress_Kiara();
       break;
   }
}


// private slots

void CCamWidget::TimeOut()
{
   int len;

   if (m_pInputBuffer == 0 || m_pImageBuffer == 0)
     return;

   len = read(m_FD, m_pInputBuffer, m_InputLen);

   if (len > 0 && m_Bandlength > 0)
   {
     switch(m_Type)
     {
       case 645:
       case 646:
         pwcx_decompress_Nala(&m_RealSize, &m_ViewSize, &m_Offset,
                              m_pInputBuffer, m_pImageBuffer,
                              PWCX_FLAG_PLANAR,
                              m_pPWCXBuffer, m_Bandlength);
         break;

       case 675:
       case 680:
       case 690:
         pwcx_decompress_Timon(&m_RealSize, &m_ViewSize, &m_Offset,
                               m_pInputBuffer, m_pImageBuffer,
                               PWCX_FLAG_PLANAR,
                               m_pPWCXBuffer, m_Bandlength);
         break;

       case 720:
       case 730:
       case 740:
       case 750:
         pwcx_decompress_Kiara(&m_RealSize, &m_ViewSize, &m_Offset,
                               m_pInputBuffer, m_pImageBuffer,
                               PWCX_FLAG_PLANAR | (m_Bayer ? PWCX_FLAG_BAYER : 0),
                               m_pPWCXBuffer, m_Bandlength);
         break;
     }
   }
   update();
}

// protected

void CCamWidget::paintEvent(QPaintEvent *e)
{
   QPainter p(this);

   p.drawImage(0, 0, m_FullImage);
   p.end();
}



// public

QSize CCamWidget::sizeHint() const
{
   return m_ImageSize;
}


void CCamWidget::SetDevice(const QString &name)
{
   m_DeviceName = name;
}


void CCamWidget::SetRaw(bool raw)
{
   m_UseRaw = raw;
}

void CCamWidget::SetBayer(bool bayer)
{
   m_Bayer = bayer;
}

bool CCamWidget::Start()
{
   struct video_capability Capability;
   struct video_picture Picture;
   struct video_window Window;
   struct pwc_probe Probe;
   struct pwc_imagesize RealSize;
   struct pwc_video_command VideoCmd;
   int i;
   QTimer *pTimer;

   qDebug(">> CCamWidget::Start()");

   qDebug("  Opening device...");
   m_FD = open(m_DeviceName.latin1(), O_RDONLY);
   if (m_FD < 0)
   {
     qWarning("Failed to open '%s', errno = %d", m_DeviceName.latin1(), errno);
     return false;
   }

   // Get capabilities
   if (ioctl(m_FD, VIDIOCGCAP, &Capability) < 0)
   {
     qDebug("ioctl(VIDIOCGCAP) failed. [%d]", errno);
     return false;
   }

   memset(&Probe, 0, sizeof(Probe));
   if (ioctl(m_FD, VIDIOCPWCPROBE, &Probe) < 0)
   {
     qDebug("PROBE failed. Not a Philips cam?");
     return false;
   }
   if (strcmp(Probe.name, Capability.name))
   {
     qDebug("Probe.name != Capability.name; not a Philips cam?");
     return false;
   }
   // We are pretty sure now we have a Philips-based cam.

   qDebug("  Picture parameters (incl. palette)...");
   if (ioctl(m_FD, VIDIOCGPICT, &Picture) < 0)
   {
     qWarning("ioctl(VIDIOCGPICT) failed. [%d]", errno);
     return false;
   }

   if (m_UseRaw)
     Picture.palette = VIDEO_PALETTE_RAW;
   else
     Picture.palette = VIDEO_PALETTE_YUV420P;

   if (ioctl(m_FD, VIDIOCSPICT, &Picture) < 0)
   {
     qWarning("ioctl(VIDIOCSPICT) failed! [%d]", errno);
     return false;
   }

   qDebug("  Window parameters (incl. size)...");
   if (ioctl(m_FD, VIDIOCGWIN, &Window) < 0)
   {
     qWarning("ioctl(VIDIOCGWIN) failed. [%d]", errno);
     return false;
   }

   if (m_Bayer)
   {
     // special raw Bayer mode
     m_ViewSize.x = 640;
     m_ViewSize.y = 480;
     m_Frames = 5;
   }

   Window.width = m_ViewSize.x;
   Window.height = m_ViewSize.y;

   Window.flags = (m_Frames << PWC_FPS_SHIFT);
   if (m_Bayer)
     Window.flags |= PWC_FPS_SNAPSHOT;

   if (ioctl(m_FD, VIDIOCSWIN, &Window) < 0)
   {
     qWarning("ioctl(VIDIOCSWIN) failed! [%d]", errno);
     return false;
   }

   m_ImageSize.setWidth(m_ViewSize.x);
   m_ImageSize.setHeight(m_ViewSize.y * 3 / 2);

   if (m_UseRaw)
   {
     if (ioctl(m_FD, VIDIOCPWCGVIDCMD, &VideoCmd) < 0)
     {
       qWarning("ioctl(VIDIOCPWCGVIDCMD) failed! [%d]", errno);
       return false;
     }

     for (i = 0; i < VideoCmd.command_len; i++)
     {
        printf("%02x ", VideoCmd.command_buf[i]);
     }
     printf("\n");
     m_Type = VideoCmd.type;
     m_InputLen = VideoCmd.frame_size;
     m_Bandlength = VideoCmd.bandlength;
     //qDebug("  Compressed image frame is %d bytes.", m_InputLen);

     if (m_Bandlength > 0)
     {
       switch(m_Type)
       {
         case 645:
         case 646:
           pwcx_init_decompress_Nala(m_Type, VideoCmd.release, &VideoCmd.command_buf, m_pPWCXBuffer);
           break;

         case 675:
         case 680:
         case 690:
           pwcx_init_decompress_Timon(m_Type, VideoCmd.release, &VideoCmd.command_buf, m_pPWCXBuffer);
           break;

         case 720:
         case 730:
         case 740:
         case 750:
           pwcx_init_decompress_Kiara(m_Type, VideoCmd.release, &VideoCmd.command_buf, m_pPWCXBuffer);
           break;

         default:
           qDebug("Unknown type of camera (%d)!", VideoCmd.type);
           break;
       } // ..switch
     } // ..m_Bandlength > 0


     if (ioctl(m_FD, VIDIOCPWCGREALSIZE, &RealSize) == 0)
     {
       m_Offset.x = (m_ViewSize.x - RealSize.width) / 2;
       m_Offset.y = (m_ViewSize.y - RealSize.height) / 2;
     }
   }

   m_ImageLen = m_ImageSize.width() * m_ImageSize.height();
   if (m_FullImage.create(m_ImageSize, 8, 256))
   {
     m_pImageBuffer = m_FullImage.bits();
     for (i = 0; i < 256; i++)
     {
        m_FullImage.setColor(i, qRgb(i, i, i)); // gray-scale
     }
   }
   else
   {
     qWarning("m_FullImage.create(..) failed!");
     return false;
   }

   /* When compressed. use own buffer, otherwise simply point input-buffer
      to image buffer.
    */
   if (m_UseRaw)
   {
     m_pInputBuffer = malloc(m_InputLen);
     if (m_pInputBuffer == 0)
     {
       qWarning("Cannot allocate input buffer.");
     }
   }
   else
   {
     if (m_pImageBuffer != 0)
     {
       m_pInputBuffer = m_pImageBuffer;
       m_InputLen = m_ImageLen;
     }
   }
   qDebug("  InputLen = %d", m_InputLen);
   updateGeometry(); // let's resize...

   qDebug("  Setting up timer...");
   pTimer = new QTimer(this, "image timer");
   connect(pTimer, SIGNAL(timeout()), this, SLOT(TimeOut()));
   pTimer->start(100);

   setBackgroundMode(Qt::NoBackground);
   qDebug("<< CCamWidget::Start()");
   return true;
}
